"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.WHITELIST = exports.STORAGE = void 0;
exports.checkPackageRemote = checkPackageRemote;
exports.cleanUpLinksRef = cleanUpLinksRef;
exports.cleanUpReadme = cleanUpReadme;
exports.generatePackageTemplate = generatePackageTemplate;
exports.generateRevision = generateRevision;
exports.getLatestReadme = getLatestReadme;
exports.hasDeprecatedVersions = hasDeprecatedVersions;
exports.hasInvalidPublishBody = hasInvalidPublishBody;
exports.isDeprecatedManifest = isDeprecatedManifest;
exports.isDifferentThanOne = isDifferentThanOne;
exports.mapManifestToSearchPackageBody = mapManifestToSearchPackageBody;
exports.mergeUplinkTimeIntoLocal = mergeUplinkTimeIntoLocal;
exports.mergeUplinkTimeIntoLocalNext = mergeUplinkTimeIntoLocalNext;
exports.mergeVersions = mergeVersions;
exports.normalizeDistTags = normalizeDistTags;
exports.normalizePackage = normalizePackage;
exports.prepareSearchPackage = prepareSearchPackage;
exports.updateUpLinkMetadata = updateUpLinkMetadata;
var _lodash = _interopRequireDefault(require("lodash"));
var _semver = _interopRequireDefault(require("semver"));
var _core = require("@verdaccio/core");
var _utils = require("@verdaccio/utils");
var _versionsUtils = require("./versions-utils");
function _interopRequireDefault(e) { return e && e.__esModule ? e : { default: e }; }
const STORAGE = exports.STORAGE = {
  PACKAGE_FILE_NAME: 'package.json',
  FILE_EXIST_ERROR: 'EEXISTS',
  NO_SUCH_FILE_ERROR: 'ENOENT',
  DEFAULT_REVISION: '0-0000000000000000'
};

/**
 * Create a new package in the storage, return a boilerplate package
 * @param name package name
 * @returns {Manifest}
 */
function generatePackageTemplate(name) {
  return {
    // standard things
    name,
    versions: {},
    time: {},
    [_core.USERS]: {},
    [_core.DIST_TAGS]: {},
    [_core.MAINTAINERS]: [],
    _uplinks: {},
    _distfiles: {},
    _attachments: {},
    _rev: ''
  };
}

/**
 * Normalize package properties, tags, revision id.
 * @param {Object} pkg package reference.
 */
function normalizePackage(pkg) {
  const pkgProperties = ['versions', 'dist-tags', '_distfiles', '_attachments', '_uplinks', 'time'];
  pkgProperties.forEach(key => {
    const pkgProp = pkg[key];
    if ((0, _utils.isNil)(pkgProp) || _core.validatioUtils.isObject(pkgProp) === false) {
      pkg[key] = {};
    }
  });
  if (typeof pkg?._rev !== 'string') {
    pkg._rev = STORAGE.DEFAULT_REVISION;
  }
  if (typeof pkg?._id !== 'string') {
    pkg._id = pkg.name;
  }

  // normalize dist-tags
  return normalizeDistTags(pkg);
}
function generateRevision(rev) {
  const _rev = rev.split('-');
  return (+_rev[0] || 0) + 1 + '-' + (0, _utils.generateRandomHexString)();
}
function getLatestReadme(pkg) {
  const versions = pkg['versions'] || {};
  const distTags = pkg[_core.DIST_TAGS] || {};
  // FIXME: here is a bit tricky add the types
  const latestVersion = distTags['latest'] ? versions[distTags['latest']] || {} : {};
  let readme = _lodash.default.trim(pkg.readme || latestVersion.readme || '');
  if (readme) {
    return readme;
  }

  // In case of empty readme - trying to get ANY readme in the following order:
  // 'next','beta','alpha','test','dev','canary'
  const readmeDistTagsPriority = ['next', 'beta', 'alpha', 'test', 'dev', 'canary'];
  readmeDistTagsPriority.forEach(function (tag) {
    if (readme) {
      return readme;
    }
    const version = distTags[tag] ? versions[distTags[tag]] || {} : {};
    readme = _lodash.default.trim(version.readme || readme);
  });
  return readme;
}

// FIXME: type any due this
function cleanUpReadme(version) {
  if ((0, _utils.isNil)(version) === false) {
    delete version.readme;
  }
  return version;
}
const WHITELIST = exports.WHITELIST = ['_rev', 'name', 'versions', 'dist-tags', 'readme', 'time', '_id', 'users', 'maintainers'];
function cleanUpLinksRef(result, keepUpLinkData) {
  const propertyToKeep = [...WHITELIST];
  if (keepUpLinkData === true) {
    propertyToKeep.push('_uplinks');
  }
  for (const i in result) {
    if (propertyToKeep.indexOf(i) === -1) {
      // Remove sections like '_uplinks' from response
      delete result[i];
    }
  }
  return result;
}

// TODO: move to abstract storage
// @deprecated use abstract.storage.ts:checkPackageRemote
function checkPackageRemote(name, isAllowPublishOffline, syncMetadata) {
  return new Promise((resolve, reject) => {
    syncMetadata(name, null, {}, (err, packageJsonLocal, upLinksErrors) => {
      // something weird
      if (err && err.status !== _core.HTTP_STATUS.NOT_FOUND) {
        return reject(err);
      }

      // checking package exist already
      if ((0, _utils.isNil)(packageJsonLocal) === false) {
        return reject(_core.errorUtils.getConflict(_core.API_ERROR.PACKAGE_EXIST));
      }
      for (let errorItem = 0; errorItem < upLinksErrors.length; errorItem++) {
        // checking error
        // if uplink fails with a status other than 404, we report failure
        if ((0, _utils.isNil)(upLinksErrors[errorItem][0]) === false) {
          if (upLinksErrors[errorItem][0].status !== _core.HTTP_STATUS.NOT_FOUND) {
            if (isAllowPublishOffline) {
              return resolve();
            }
            return reject(_core.errorUtils.getServiceUnavailable(_core.API_ERROR.UPLINK_OFFLINE_PUBLISH));
          }
        }
      }
      return resolve();
    });
  });
}

// @deprecated use export function mergeUplinkTimeIntoLocalNext
function mergeUplinkTimeIntoLocal(localMetadata, remoteMetadata) {
  if ('time' in remoteMetadata) {
    return Object.assign({}, localMetadata.time, remoteMetadata.time);
  }
  return localMetadata.time;
}
function mergeUplinkTimeIntoLocalNext(cacheManifest, remoteManifest) {
  if ('time' in remoteManifest) {
    // remote override cache time conflicts
    return {
      ...cacheManifest,
      time: {
        ...cacheManifest.time,
        ...remoteManifest.time
      }
    };
  }
  return cacheManifest;
}
function updateUpLinkMetadata(uplinkId, manifest, etag) {
  const _uplinks = {
    ...manifest._uplinks,
    [uplinkId]: {
      etag,
      fetched: Date.now()
    }
  };
  return {
    ...manifest,
    _uplinks
  };
}
function prepareSearchPackage(data) {
  const latest = _core.pkgUtils.getLatest(data);
  if (latest && data.versions[latest]) {
    const version = data.versions[latest];
    const versions = {
      [latest]: 'latest'
    };
    const pkg = {
      name: version.name,
      description: version.description,
      [_core.DIST_TAGS]: {
        latest
      },
      maintainers: version.maintainers || [version.author].filter(Boolean),
      author: version.author,
      repository: version.repository,
      readmeFilename: version.readmeFilename || '',
      homepage: version.homepage,
      keywords: version.keywords,
      bugs: version.bugs,
      license: version.license,
      // time: {
      //   modified: time,
      // },
      versions
    };
    return pkg;
  }
}
function isDifferentThanOne(versions) {
  return Object.keys(versions).length !== 1;
}

// @deprecated use validationUtils.validatePublishNewPackage
function hasInvalidPublishBody(manifest) {
  if (!manifest) {
    return false;
  }
  const {
    _attachments,
    versions
  } = manifest;
  const res = (0, _utils.isObject)(_attachments) === false || isDifferentThanOne(_attachments) || (0, _utils.isObject)(versions) === false || isDifferentThanOne(versions);
  return res;
}

/**
 * Function gets a local info and an info from uplinks and tries to merge it
 exported for unit tests only.
  * @param {*} local
  * @param {*} remoteManifest
  * @param {*} config configuration file
  */
function mergeVersions(cacheManifest, remoteManifest) {
  let _cacheManifest = {
    ...cacheManifest
  };
  const {
    versions,
    time
  } = remoteManifest;
  // copy new versions to a cache
  // NOTE: if a certain version was updated, we can't refresh it reliably
  for (const i in versions) {
    if (typeof cacheManifest.versions[i] === 'undefined') {
      _cacheManifest.versions[i] = versions[i];
    }
  }
  for (const i in time) {
    if (typeof cacheManifest.time[i] === 'undefined') {
      _cacheManifest.time[i] = time[i];
    }
  }
  for (const distTag in remoteManifest[_core.DIST_TAGS]) {
    if (_cacheManifest[_core.DIST_TAGS][distTag] !== remoteManifest[_core.DIST_TAGS][distTag]) {
      if (!_cacheManifest[_core.DIST_TAGS][distTag] || _semver.default.lte(_cacheManifest[_core.DIST_TAGS][distTag], remoteManifest[_core.DIST_TAGS][distTag])) {
        _cacheManifest[_core.DIST_TAGS][distTag] = remoteManifest[_core.DIST_TAGS][distTag];
      }
      if (distTag === 'latest' && _cacheManifest[_core.DIST_TAGS][distTag] === remoteManifest[_core.DIST_TAGS][distTag]) {
        // NOTE: this override the latest publish readme from local cache with
        // the remote one
        cacheManifest = {
          ..._cacheManifest,
          readme: remoteManifest.readme
        };
      }
    }
  }

  // TODO: Should we merge owners? _cacheManifest[MAINTAINERS]

  return cacheManifest;
}

/**
 * Normalize dist-tags.
 *
 * There is a legacy behaviour where the dist-tags could be an array, in such
 * case, the array is orderded and the highest version becames the
 * latest.
 *
 * The dist-tag tags must be plain strigs, if the latest is empty (for whatever reason) is
 * normalized to be the highest version available.
 *
 * This function cleans up every invalid version on dist-tags, but does not remove
 * invalid versions from the manifest.
 *
 * @param {*} data
 */
function normalizeDistTags(manifest) {
  let sorted;
  // handle missing latest dist-tag
  if (!manifest[_core.DIST_TAGS].latest) {
    // if there is no latest tag, set the highest known version based on semver sort
    sorted = (0, _versionsUtils.sortVersionsAndFilterInvalid)(Object.keys(manifest.versions));
    if (sorted?.length) {
      // get the highest published version
      manifest[_core.DIST_TAGS].latest = sorted.pop();
    }
  }
  for (const tag in manifest[_core.DIST_TAGS]) {
    // deprecated (will be removed un future majors)
    // this should not happen, tags should be plain strings, legacy fallback
    if (_lodash.default.isArray(manifest[_core.DIST_TAGS][tag])) {
      if (manifest[_core.DIST_TAGS][tag].length) {
        // sort array
        // FIXME: this is clearly wrong, we need to research why this is like this.
        // @ts-ignore
        sorted = (0, _versionsUtils.sortVersionsAndFilterInvalid)(manifest[_core.DIST_TAGS][tag]);
        if (sorted.length) {
          // use highest version based on semver sort
          manifest[_core.DIST_TAGS][tag] = sorted.pop();
        }
      } else {
        delete manifest[_core.DIST_TAGS][tag];
      }
    } else if (_lodash.default.isString(manifest[_core.DIST_TAGS][tag])) {
      if (!_semver.default.parse(manifest[_core.DIST_TAGS][tag], true)) {
        // if the version is invalid, delete the dist-tag entry
        delete manifest[_core.DIST_TAGS][tag];
      }
    }
  }
  return manifest;
}
function hasDeprecatedVersions(pkgInfo) {
  const {
    versions
  } = pkgInfo;
  for (const version in versions) {
    if (Object.prototype.hasOwnProperty.call(versions[version], 'deprecated')) {
      return true;
    }
  }
  return false;
}
function isDeprecatedManifest(manifest) {
  return hasDeprecatedVersions(manifest) && Object.keys(manifest._attachments).length === 0;
}
function mapManifestToSearchPackageBody(pkg, searchItem) {
  const latest = _core.pkgUtils.getLatest(pkg);
  const version = pkg.versions[latest];
  const result = {
    name: version.name,
    scope: '',
    description: version.description,
    version: latest,
    keywords: version.keywords,
    date: pkg.time[latest],
    // FIXME: type
    author: version.author,
    // FIXME: not possible fill this out from a private package
    publisher: {},
    // FIXME: type
    maintainers: version.maintainers,
    links: {
      npm: '',
      homepage: version.homepage,
      repository: version.repository,
      bugs: version.bugs
    }
  };
  if (typeof searchItem.package.scoped === 'string') {
    result.scope = searchItem.package.scoped;
  }
  return result;
}
//# sourceMappingURL=storage-utils.js.map