"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.getVersion = getVersion;
exports.isNewerVersion = isNewerVersion;
exports.removeLowerVersions = removeLowerVersions;
exports.sortVersionsAndFilterInvalid = sortVersionsAndFilterInvalid;
exports.tagVersion = tagVersion;
exports.tagVersionNext = tagVersionNext;
var _debug = _interopRequireDefault(require("debug"));
var _lodash = _interopRequireDefault(require("lodash"));
var _semver = _interopRequireDefault(require("semver"));
var _core = require("@verdaccio/core");
function _interopRequireDefault(e) { return e && e.__esModule ? e : { default: e }; }
const debug = (0, _debug.default)('verdaccio:storage:utils');

/**
 * Gets version from a package object taking into account semver weirdness.
 * @return {String} return the semantic version of a package
 */
function getVersion(versions, version) {
  if (!versions) {
    return;
  }

  // this condition must allow cast
  if (_lodash.default.isNil(versions[version]) === false) {
    return versions[version];
  }
  const versionSemver = _semver.default.parse(version, true);
  if (versionSemver === null) {
    return;
  }
  for (const versionItem in versions) {
    if (Object.prototype.hasOwnProperty.call(versions, versionItem)) {
      // @ts-ignore
      if (versionSemver.compare(_semver.default.parse(versionItem, true)) === 0) {
        return versions[versionItem];
      }
    }
  }
}

/**
 * Function filters out bad semver versions and sorts the array.
 * @return {Array} sorted Array
 */
function sortVersionsAndFilterInvalid(listVersions) {
  return listVersions.filter(function (version) {
    if (!_semver.default.parse(version, true)) {
      return false;
    }
    return true;
  })
  // FIXME: it seems the @types/semver do not handle a legitimate method named 'compareLoose'
  // @ts-ignore
  .sort(_semver.default.compareLoose).map(String);
}

/**
 * Create a tag for a package
 * @param {*} data
 * @param {*} version
 * @param {*} tag
 * @return {Boolean} whether a package has been tagged
 * @deprecated
 */
function tagVersion(data, version, tag) {
  if (tag && data[_core.DIST_TAGS][tag] !== version && _semver.default.parse(version, true)) {
    // valid version - store
    data[_core.DIST_TAGS][tag] = version;
    return true;
  }
  return false;
}

/**
 *
 * @param manifest
 * @param version
 * @param tag
 * @returns
 */
function tagVersionNext(manifest, version, tag) {
  const data = {
    ...manifest
  };
  if (tag && data[_core.DIST_TAGS][tag] !== version && _semver.default.parse(version, true)) {
    // valid version - store
    data[_core.DIST_TAGS][tag] = version;
  }
  return data;
}

/**
 *  Check if the version is newer than the older version.
 * @param newVersion
 * @param oldVersion
 * @returns
 */
function isNewerVersion(newVersion, oldVersion) {
  const comparisonResult = _semver.default.compare(newVersion, oldVersion);
  return comparisonResult === 1 || comparisonResult === 0;
}

/**
 * Remove duplicates from search results.
 * @param {Array} objects
 * @return {Array} filtered array
 */
function removeLowerVersions(objects) {
  const versionMap = new Map();

  // Iterate through the array and keep the highest version for each name
  objects.forEach(item => {
    const {
      name,
      version
    } = item.package;
    const key = name;
    if (versionMap.has(name) === false || isNewerVersion(version, versionMap.get(name))) {
      debug('keeping %o@%o', name, version);
      versionMap.set(key, version);
    }
  });

  // Filter objects based on the version map
  return objects.reduce((acc, item) => {
    const {
      name,
      version
    } = item.package;
    if (versionMap.has(name) && versionMap.get(name) === version && acc.find(i => i.package.name === name) === undefined) {
      debug('adding %o@%o', name, version);
      acc.push(item);
    }
    return acc;
  }, []);
}
//# sourceMappingURL=versions-utils.js.map