"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.resourceNotAvailable = exports.noSuchFile = exports.Storage = exports.PROTO_NAME = void 0;
var _assert = _interopRequireDefault(require("assert"));
var _debug = _interopRequireDefault(require("debug"));
var _lodash = _interopRequireWildcard(require("lodash"));
var _path = require("path");
var _stream = require("stream");
var _promises = require("stream/promises");
var _url = _interopRequireDefault(require("url"));
var _config = require("@verdaccio/config");
var _core = require("@verdaccio/core");
var _loaders = require("@verdaccio/loaders");
var _proxy = require("@verdaccio/proxy");
var _search = _interopRequireDefault(require("@verdaccio/search"));
var _tarball = require("@verdaccio/tarball");
var _utils = require("@verdaccio/utils");
var _2 = require(".");
var _starUtils = require("./lib/star-utils");
var _storageUtils = require("./lib/storage-utils");
var _versionsUtils = require("./lib/versions-utils");
var _localStorage = require("./local-storage");
function _getRequireWildcardCache(e) { if ("function" != typeof WeakMap) return null; var r = new WeakMap(), t = new WeakMap(); return (_getRequireWildcardCache = function (e) { return e ? t : r; })(e); }
function _interopRequireWildcard(e, r) { if (!r && e && e.__esModule) return e; if (null === e || "object" != typeof e && "function" != typeof e) return { default: e }; var t = _getRequireWildcardCache(r); if (t && t.has(e)) return t.get(e); var n = { __proto__: null }, a = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var u in e) if ("default" !== u && {}.hasOwnProperty.call(e, u)) { var i = a ? Object.getOwnPropertyDescriptor(e, u) : null; i && (i.get || i.set) ? Object.defineProperty(n, u, i) : n[u] = e[u]; } return n.default = e, t && t.set(e, n), n; }
function _interopRequireDefault(e) { return e && e.__esModule ? e : { default: e }; }
const debug = (0, _debug.default)('verdaccio:storage');
const noSuchFile = exports.noSuchFile = 'ENOENT';
const resourceNotAvailable = exports.resourceNotAvailable = 'EAGAIN';
const PROTO_NAME = exports.PROTO_NAME = '__proto__';
class Storage {
  constructor(config, logger) {
    this.config = config;
    this.logger = logger.child({
      module: 'storage'
    });
    this.uplinks = (0, _proxy.setupUpLinks)(config, this.logger);
    this.searchService = new _search.default(config, this.logger);
    this.filters = null;
    // @ts-ignore
    this.localStorage = null;
    debug('uplinks available %o', Object.keys(this.uplinks));
  }
  static ABBREVIATED_HEADER = 'application/vnd.npm.install-v1+json';

  /**
   * Change an existing package (i.e. unpublish one version)
   Function changes a package info from local storage and all uplinks with write access./
   Used storages: local (write)
   */
  async changePackage(name, metadata, revision) {
    debug('change existing package for package %o revision %o', name, revision);
    debug(`change manifest tags for %o revision %o`, name, revision);
    if (!_core.validatioUtils.isObject(metadata.versions) || !_core.validatioUtils.isObject(metadata[_core.DIST_TAGS])) {
      debug(`change manifest bad data for %o`, name);
      throw _core.errorUtils.getBadData();
    }
    debug(`change manifest updating manifest for %o`, name);
    await this.updatePackage(name, async localData => {
      // eslint-disable-next-line guard-for-in
      for (const version in localData.versions) {
        const incomingVersion = metadata.versions[version];
        if (_lodash.default.isNil(incomingVersion)) {
          this.logger.info({
            name: name,
            version: version
          }, 'unpublishing @{name}@@{version}');

          // FIXME: I prefer return a new object rather mutate the metadata
          delete localData.versions[version];
          delete localData.time[version];
          for (const file in localData._attachments) {
            if (localData._attachments[file].version === version) {
              delete localData._attachments[file].version;
            }
          }
        } else if (Object.prototype.hasOwnProperty.call(incomingVersion, 'deprecated')) {
          const incomingDeprecated = incomingVersion.deprecated;
          if (incomingDeprecated != localData.versions[version].deprecated) {
            if (!incomingDeprecated) {
              this.logger.info({
                name: name,
                version: version
              }, 'undeprecating @{name}@@{version}');
              delete localData.versions[version].deprecated;
            } else {
              this.logger.info({
                name: name,
                version: version
              }, 'deprecating @{name}@@{version}');
              localData.versions[version].deprecated = incomingDeprecated;
            }
            localData.time.modified = new Date().toISOString();
          }
        }
      }
      localData[_core.USERS] = metadata[_core.USERS];
      localData[_core.DIST_TAGS] = metadata[_core.DIST_TAGS];
      localData[_core.MAINTAINERS] = metadata[_core.MAINTAINERS];
      return localData;
    });
  }
  async removePackage(name, revision, username) {
    debug('remove package %o', name);
    await this.removePackageByRevision(name, revision, username);
  }

  /**
   Remove a tarball from a system
   Function removes a tarball from local storage.
   Tarball in question should not be linked to in any existing
   versions, i.e. package version should be unpublished first.
   Used storage: local (write)
   */
  async removeTarball(name, filename,
  // eslint-disable-next-line @typescript-eslint/no-unused-vars
  _revision, username) {
    debug('remove tarball %s for %s', filename, name);
    (0, _assert.default)(_core.validatioUtils.validateName(filename));
    const storage = this.getPrivatePackageStorage(name);
    if (!storage) {
      debug(`method not implemented for storage`);
      this.logger.error('method for remove tarball not implemented');
      throw _core.errorUtils.getInternalError(_core.API_ERROR.INTERNAL_SERVER_ERROR);
    }
    try {
      const cacheManifest = await storage.readPackage(name);
      if (!cacheManifest._attachments[filename]) {
        throw _core.errorUtils.getNotFound('no such file available');
      }

      // check if logged in user is allowed to remove tarball
      await this.checkAllowedToChangePackage(cacheManifest, username);
    } catch (err) {
      if (err.code === noSuchFile) {
        throw _core.errorUtils.getNotFound();
      }
      throw err;
    }
    const manifest = await this.updatePackage(name, async data => {
      let newData = {
        ...data
      };
      delete data._attachments[filename];
      return newData;
    });
    try {
      const storage = this.getPrivatePackageStorage(name);
      if (!storage) {
        debug(`method not implemented for storage`);
        this.logger.error('method for remove tarball not implemented');
        throw _core.errorUtils.getInternalError(_core.API_ERROR.INTERNAL_SERVER_ERROR);
      }
      await storage.deletePackage(filename);
      debug('package %s removed', filename);
    } catch (err) {
      this.logger.error({
        err
      }, 'error removing %s from storage');
      throw err;
    }
    return manifest;
  }

  /**
   * Handle search on packages and proxies.
   * Iterate all proxies configured and search in all endpoints in v2 and pipe all responses
   *  once the proxies request has finished search in local storage for all packages
   * (privated and cached).
   */
  async search(options) {
    debug('search on cache packages');
    const cachePackages = await this.getCachedPackages(options.query);
    debug('search found on cache packages %o', cachePackages.length);
    const remotePackages = await this.searchService.search(options);
    debug('search found on remote packages %o', remotePackages.length);
    const totalResults = [...cachePackages, ...remotePackages];
    const uniqueResults = (0, _versionsUtils.removeLowerVersions)(totalResults);
    debug('unique results %o', uniqueResults.length);
    return uniqueResults;
  }
  async getTarballFromUpstream(name, filename, {
    signal
  }) {
    this.logger.info({
      name,
      filename
    }, 'get tarball for package @{name} filename @{filename} from uplink');
    let cachedManifest = null;
    try {
      cachedManifest = await this.getPackageLocalMetadata(name);
    } catch (err) {
      debug('error on get package local metadata %o', err);
    }
    // dist url should be on local cache metadata
    if (cachedManifest?._distfiles && typeof cachedManifest?._distfiles[filename]?.url === 'string') {
      debug('dist file found, using it %o', cachedManifest?._distfiles[filename].url);
      // dist file found, proceed to download
      const distFile = cachedManifest._distfiles[filename];
      let current_length = 0;
      let expected_length;
      const passThroughRemoteStream = new _stream.PassThrough();
      const proxy = this.getUpLinkForDistFile(name, distFile);
      const remoteStream = proxy.fetchTarball(distFile.url, {});
      remoteStream.on('request', async () => {
        try {
          debug('remote stream request');
          const storage = this.getPrivatePackageStorage(name);
          if (proxy.config.cache === true && storage) {
            const localStorageWriteStream = await storage.writeTarball(filename, {
              signal
            });
            await (0, _promises.pipeline)(remoteStream, passThroughRemoteStream, localStorageWriteStream, {
              signal
            });
          } else {
            await (0, _promises.pipeline)(remoteStream, passThroughRemoteStream, {
              signal
            });
          }
        } catch (err) {
          debug('error on pipeline downloading tarball for package %o', name);
          passThroughRemoteStream.emit('error', err);
        }
      });
      remoteStream.on('response', async res => {
        if (res.statusCode === _core.HTTP_STATUS.NOT_FOUND) {
          debug('remote stream response 404');
          passThroughRemoteStream.emit('error', _core.errorUtils.getNotFound(_core.errorUtils.API_ERROR.NOT_FILE_UPLINK));
          return;
        }
        if (!(res.statusCode >= _core.HTTP_STATUS.OK && res.statusCode < _core.HTTP_STATUS.MULTIPLE_CHOICES)) {
          debug('remote stream response ok');
          passThroughRemoteStream.emit('error', _core.errorUtils.getInternalError(`bad uplink status code: ${res.statusCode}`));
          return;
        }
        if (res.headers[_core.HEADER_TYPE.CONTENT_LENGTH]) {
          expected_length = res.headers[_core.HEADER_TYPE.CONTENT_LENGTH];
          debug('remote stream response content length %o', expected_length);
          passThroughRemoteStream.emit(_core.HEADER_TYPE.CONTENT_LENGTH, res.headers[_core.HEADER_TYPE.CONTENT_LENGTH]);
        }
      }).on('downloadProgress', progress => {
        current_length = progress.transferred;
        if (typeof expected_length === 'undefined' && progress.total) {
          expected_length = progress.total;
        }
      }).on('end', () => {
        if (expected_length && current_length != expected_length) {
          debug('stream end, but length mismatch %o %o', current_length, expected_length);
          passThroughRemoteStream.emit('error', _core.errorUtils.getInternalError(_core.API_ERROR.CONTENT_MISMATCH));
        }
        debug('remote stream end');
      }).on('error', err => {
        debug('remote stream error %o', err);
        passThroughRemoteStream.emit('error', err);
      });
      return passThroughRemoteStream;
    } else {
      debug('dist file not found, proceed update upstream');
      // no dist url found, proceed to fetch from upstream
      // should not be the case
      const passThroughRemoteStream = new _stream.PassThrough();
      // ensure get the latest data
      const [updatedManifest] = await this.syncUplinksMetadata(name, cachedManifest, {
        uplinksLook: true
      });
      const distFile = updatedManifest._distfiles[filename];
      if (updatedManifest === null || !distFile) {
        debug('remote tarball not found');
        throw _core.errorUtils.getNotFound(_core.API_ERROR.NO_SUCH_FILE);
      }
      const proxy = this.getUpLinkForDistFile(name, distFile);
      const remoteStream = proxy.fetchTarball(distFile.url, {});
      remoteStream.on('response', async () => {
        try {
          const storage = this.getPrivatePackageStorage(name);
          if (proxy.config.cache === true && storage) {
            debug('cache remote tarball enabled');
            const localStorageWriteStream = await storage.writeTarball(filename, {
              signal
            });
            await (0, _promises.pipeline)(remoteStream, passThroughRemoteStream, localStorageWriteStream, {
              signal
            });
          } else {
            debug('cache remote tarball disabled');
            await (0, _promises.pipeline)(remoteStream, passThroughRemoteStream, {
              signal
            });
          }
        } catch (err) {
          debug('error on pipeline downloading tarball for package %o', name);
          passThroughRemoteStream.emit('error', err);
        }
      });
      return passThroughRemoteStream;
    }
  }

  /**
   *
   * @param name
   * @param filename
   * @param param2
   * @returns
   */
  async getTarball(name, filename, {
    signal
  }) {
    this.logger.info({
      name,
      filename
    }, 'get tarball for package @{name} filename @{filename}');
    debug('get tarball for package %o filename %o', name, filename);
    // TODO: check if isOpen is need it after all.
    // eslint-disable-next-line @typescript-eslint/no-unused-vars
    let isOpen = false;
    const localTarballStream = new _stream.PassThrough();
    const localStream = await this.getLocalTarball(name, filename, {
      signal
    });
    localStream.on('open', async () => {
      // eslint-disable-next-line @typescript-eslint/no-unused-vars
      isOpen = true;
      await (0, _promises.pipeline)(localStream, localTarballStream, {
        signal
      });
    });
    localStream.on('error', err => {
      // eslint-disable-next-line no-console
      if (err.code === _storageUtils.STORAGE.NO_SUCH_FILE_ERROR || err.code === _core.HTTP_STATUS.NOT_FOUND) {
        this.getTarballFromUpstream(name, filename, {
          signal
        }).then(uplinkStream => {
          (0, _promises.pipeline)(uplinkStream, localTarballStream, {
            signal
          }).then(() => {
            debug('successfully downloaded tarball for package %o filename %o', name, filename);
          }).catch(err => {
            localTarballStream.emit('error', err);
          });
        }).catch(err => {
          localTarballStream.emit('error', err);
        });
      } else {
        this.logger.error({
          err: err.message
        }, 'some error on fatal @{err}');
        localTarballStream.emit('error', err);
      }
    });
    return localTarballStream;
  }
  async getPackageByVersion(options) {
    const queryVersion = options.version;
    if (_lodash.default.isNil(queryVersion)) {
      throw _core.errorUtils.getNotFound(`${_core.API_ERROR.VERSION_NOT_EXIST}: ${queryVersion}`);
    }

    // we have version, so we need to return specific version
    const [convertedManifest] = await this.getPackageNext(options);
    const version = (0, _versionsUtils.getVersion)(convertedManifest.versions, queryVersion);
    debug('query by latest version %o and result %o', queryVersion, version);
    if (typeof version !== 'undefined') {
      debug('latest version found %o', version);
      return (0, _tarball.convertDistVersionToLocalTarballsUrl)(convertedManifest.name, version, options.requestOptions, this.config.url_prefix);
    }

    // the version could be a dist-tag eg: beta, alpha, so we find the matched version
    // on disg-tag list
    if (_lodash.default.isNil(convertedManifest[_core.DIST_TAGS]) === false) {
      if (_lodash.default.isNil(convertedManifest[_core.DIST_TAGS][queryVersion]) === false) {
        // the version found as a distag
        const matchedDisTagVersion = convertedManifest[_core.DIST_TAGS][queryVersion];
        debug('dist-tag version found %o', matchedDisTagVersion);
        const disTagVersion = (0, _versionsUtils.getVersion)(convertedManifest.versions, matchedDisTagVersion);
        if (typeof disTagVersion !== 'undefined') {
          debug('dist-tag found %o', disTagVersion);
          return (0, _tarball.convertDistVersionToLocalTarballsUrl)(convertedManifest.name, disTagVersion, options.requestOptions, this.config.url_prefix);
        }
      }
    } else {
      debug('dist tag not detected');
    }

    // we didn't find the version, not found error
    debug('package version not found %o', queryVersion);
    throw _core.errorUtils.getNotFound(`${_core.API_ERROR.VERSION_NOT_EXIST}: ${queryVersion}`);
  }
  async getPackageManifest(options) {
    // convert dist remotes to local bars
    const [manifest] = await this.getPackageNext(options);

    // If change access is requested (?write=true), then check if logged in user is allowed to change package
    if (options.byPassCache === true) {
      try {
        await this.checkAllowedToChangePackage(manifest, options.requestOptions.username);
      } catch (error) {
        this.logger.error({
          error: error.message
        }, 'getting package has failed: @{error}');
        throw _core.errorUtils.getBadRequest(error.message);
      }
    }
    const convertedManifest = (0, _tarball.convertDistRemoteToLocalTarballUrls)(manifest, options.requestOptions, this.config.url_prefix);
    return convertedManifest;
  }
  convertAbbreviatedManifest(manifest) {
    const abbreviatedVersions = Object.keys(manifest.versions).reduce((acc, version) => {
      const _version = manifest.versions[version];
      // This should be align with this document
      // https://github.com/npm/registry/blob/master/docs/responses/package-metadata.md#abbreviated-version-object
      const _version_abbreviated = {
        name: _version.name,
        version: _version.version,
        deprecated: _version.deprecated,
        bin: _version.bin,
        dist: _version.dist,
        engines: _version.engines,
        funding: _version.funding,
        directories: _version.directories,
        dependencies: _version.dependencies,
        devDependencies: _version.devDependencies,
        peerDependencies: _version.peerDependencies,
        optionalDependencies: _version.optionalDependencies,
        bundleDependencies: _version.bundleDependencies,
        cpu: _version.cpu,
        os: _version.os,
        peerDependenciesMeta: _version.peerDependenciesMeta,
        acceptDependencies: _version.acceptDependencies,
        // npm cli specifics
        _hasShrinkwrap: _version._hasShrinkwrap,
        hasInstallScript: _version.hasInstallScript
      };
      acc[version] = _version_abbreviated;
      return acc;
    }, {});
    const convertedManifest = {
      name: manifest['name'],
      [_core.DIST_TAGS]: manifest[_core.DIST_TAGS],
      versions: abbreviatedVersions,
      modified: manifest.time.modified,
      // NOTE: special case for pnpm https://github.com/pnpm/rfcs/pull/2
      time: manifest.time
    };
    return convertedManifest;
  }

  /**
   * Return a manifest or version based on the options.
   * @param options {Object}
   * @returns A package manifest or specific version
   */
  async getPackageByOptions(options) {
    // if no version we return the whole manifest
    if (_lodash.default.isNil(options.version) === false) {
      return this.getPackageByVersion(options);
    } else {
      const manifest = await this.getPackageManifest(options);
      if (options.abbreviated === true) {
        debug('abbreviated manifest');
        return this.convertAbbreviatedManifest(manifest);
      }
      return manifest;
    }
  }
  async getLocalDatabase() {
    debug('get local database');
    const storage = this.localStorage.getStoragePlugin();
    const database = await storage.get();
    const packages = [];
    for (const pkg of database) {
      debug('get local database %o', pkg);
      const manifest = await this.getPackageLocalMetadata(pkg);
      const latest = manifest[_core.DIST_TAGS].latest;
      if (latest && manifest.versions[latest]) {
        const version = manifest.versions[latest];
        const timeList = manifest.time;
        const time = timeList[latest];
        // @ts-ignore
        version.time = time;

        // Add for stars api
        // @ts-ignore
        version.users = manifest.users;
        packages.push(version);
      } else {
        this.logger.warn({
          package: pkg
        }, 'package @{package} does not have a "latest" tag?');
      }
    }
    return packages;
  }
  saveToken(token) {
    if (_lodash.default.isFunction(this.localStorage.getStoragePlugin().saveToken) === false) {
      return Promise.reject(_core.errorUtils.getCode(_core.HTTP_STATUS.SERVICE_UNAVAILABLE, _core.SUPPORT_ERRORS.PLUGIN_MISSING_INTERFACE));
    }
    return this.localStorage.getStoragePlugin().saveToken(token);
  }
  deleteToken(user, tokenKey) {
    if (_lodash.default.isFunction(this.localStorage.getStoragePlugin().deleteToken) === false) {
      return Promise.reject(_core.errorUtils.getCode(_core.HTTP_STATUS.SERVICE_UNAVAILABLE, _core.SUPPORT_ERRORS.PLUGIN_MISSING_INTERFACE));
    }
    return this.localStorage.getStoragePlugin().deleteToken(user, tokenKey);
  }
  readTokens(filter) {
    if (_lodash.default.isFunction(this.localStorage.getStoragePlugin().readTokens) === false) {
      return Promise.reject(_core.errorUtils.getCode(_core.HTTP_STATUS.SERVICE_UNAVAILABLE, _core.SUPPORT_ERRORS.PLUGIN_MISSING_INTERFACE));
    }
    return this.localStorage.getStoragePlugin().readTokens(filter);
  }

  /**
   * Initialize the storage asynchronously.
   * @param config Config
   * @param filters Filters
   * @returns Storage instance
   */
  async init(config) {
    if (this.localStorage === null) {
      this.localStorage = new _localStorage.LocalStorage(this.config, this.logger);
      await this.localStorage.init();
      debug('local init storage initialized');
      await this.localStorage.getSecret(config);
      debug('local storage secret initialized');
    } else {
      debug('storage has been already initialized');
    }
    if (!this.filters) {
      this.filters = await (0, _loaders.asyncLoadPlugin)(this.config.filters, {
        config: this.config,
        logger: this.logger
      }, plugin => {
        return typeof plugin.filter_metadata !== 'undefined';
      }, this.config?.serverSettings?.pluginPrefix, _core.PLUGIN_CATEGORY.FILTER);
      debug('filters available %o', this.filters.length);
    }
    return;
  }

  /**
   * Retrieve a wrapper that provide access to the package location.
   * @param {Object} pkgName package name.
   * @return {Object}
   */
  getPrivatePackageStorage(pkgName) {
    debug('get local storage for %o', pkgName);
    return this.localStorage.getStoragePlugin().getPackageStorage(pkgName);
  }

  /**
   * Create a tarball stream from a package.
   * @param name
   * @param filename
   * @param options
   * @returns
   */
  async getLocalTarball(pkgName, filename, {
    signal
  }) {
    (0, _assert.default)(_core.validatioUtils.validateName(filename));
    const storage = this.getPrivatePackageStorage(pkgName);
    if (typeof storage === 'undefined') {
      return this.createFailureStreamResponseNext();
    }
    return await storage.readTarball(filename, {
      signal
    });
  }
  async getCachedPackages(query) {
    debug('search on each package', query);
    const results = [];
    if (typeof query === 'undefined' || typeof query?.text === 'undefined') {
      debug('search query for cached not found');
      return results;
    }
    this.logger.info({
      t: query.text,
      q: query.quality,
      p: query.popularity,
      m: query.maintenance,
      s: query.size
    }, 'search by text @{t}| maintenance @{m}| quality @{q}| popularity @{p}');
    if (typeof this.localStorage.getStoragePlugin().search === 'undefined') {
      this.logger.info('plugin search not implemented yet');
    } else {
      debug('search on each package by plugin query');
      const items = await this.localStorage.getStoragePlugin().search(query);
      try {
        for (const searchItem of items) {
          const manifest = await this.getPackageLocalMetadata(searchItem.package.name);
          if (_lodash.default.isEmpty(manifest?.versions) === false) {
            const searchPackage = (0, _storageUtils.mapManifestToSearchPackageBody)(manifest, searchItem);
            debug('search local stream found %o', searchPackage.name);
            const searchPackageItem = {
              package: searchPackage,
              score: searchItem.score,
              verdaccioPkgCached: searchItem.verdaccioPkgCached,
              verdaccioPrivate: searchItem.verdaccioPrivate,
              flags: searchItem?.flags,
              // FUTURE: find a better way to calculate the score
              searchScore: 1
            };
            results.push(searchPackageItem);
          } else {
            debug('local item without versions detected %s', searchItem.package.name);
          }
        }
        debug('search local stream end');
      } catch (err) {
        this.logger.error({
          err,
          query
        }, 'error on search by plugin @{err.message}');
        throw err;
      }
    }
    return results;
  }
  async removePackageByRevision(pkgName, revision, username) {
    const storage = this.getPrivatePackageStorage(pkgName);
    debug('get package metadata for %o', pkgName);
    if (typeof storage === 'undefined') {
      throw _core.errorUtils.getServiceUnavailable('storage not initialized');
    }
    let manifest;
    try {
      manifest = await storage.readPackage(pkgName);
      manifest = (0, _storageUtils.normalizePackage)(manifest);
    } catch (err) {
      if (err.code === _storageUtils.STORAGE.NO_SUCH_FILE_ERROR || err.code === _core.HTTP_STATUS.NOT_FOUND) {
        this.logger.info({
          pkgName,
          revision
        }, 'package not found');
        throw _core.errorUtils.getNotFound();
      }
      this.logger.error({
        pkgName,
        revision,
        err: err.message
      }, 'error @{err} while reading package @{pkgName}-{revision}');
      throw err;
    }

    // TODO:  move this to another method
    try {
      // check if logged in user is allowed to remove package
      await this.checkAllowedToChangePackage(manifest, username);
      await this.localStorage.getStoragePlugin().remove(pkgName);
      // remove each attachment
      const attachments = Object.keys(manifest._attachments);
      debug('attachments to remove %s', attachments?.length);
      for (let attachment of attachments) {
        debug('remove attachment %s', attachment);
        await storage.deletePackage(attachment);
        this.logger.info({
          attachment
        }, 'attachment @{attachment} removed');
      }
      // remove package.json
      debug('remove package.json');
      await storage.deletePackage(_storageUtils.STORAGE.PACKAGE_FILE_NAME);
      // remove folder
      debug('remove package folder');
      await storage.removePackage();
      this.logger.info({
        pkgName
      }, 'package @{pkgName} removed');
    } catch (err) {
      this.logger.error({
        err
      }, 'removed package has failed @{err.message}');
      throw _core.errorUtils.getBadData(err.message);
    }
  }

  /**
   * Get a package local manifest.
   *
   * Fails if package is not found.
   * @param name package name
   * @param revision of package
   * @returns local manifest
   */
  // eslint-disable-next-line @typescript-eslint/no-unused-vars
  async getPackageLocalMetadata(name, _revision) {
    const storage = this.getPrivatePackageStorage(name);
    debug('get package metadata for %o', name);
    if (typeof storage === 'undefined') {
      throw _core.errorUtils.getServiceUnavailable('storage not initialized');
    }
    try {
      const result = await storage.readPackage(name);
      return (0, _storageUtils.normalizePackage)(result);
    } catch (err) {
      if (err.code === _storageUtils.STORAGE.NO_SUCH_FILE_ERROR || err.code === _core.HTTP_STATUS.NOT_FOUND) {
        debug('package %s not found', name);
        throw _core.errorUtils.getNotFound();
      }
      this.logger.error({
        err: err,
        file: _storageUtils.STORAGE.PACKAGE_FILE_NAME
      }, `error reading  @{file}: @{!err.message}`);
      throw _core.errorUtils.getInternalError();
    }
  }

  /**
   * Fail the stream response with an not found error.
   * @returns
   */
  createFailureStreamResponseNext() {
    const stream = new _stream.PassThrough();

    // we ensure fails on the next tick into the event loop
    process.nextTick(() => {
      stream.emit('error', _core.errorUtils.getNotFound(_core.API_ERROR.NO_SUCH_FILE));
    });
    return stream;
  }

  /**
   * Update a package and merge tags
   * @param name package name
   * @param tags list of dist-tags
   */
  async mergeTagsNext(name, tags) {
    return await this.updatePackage(name, async data => {
      let newData = {
        ...data
      };
      for (const tag of Object.keys(tags)) {
        // this handle dist-tag rm command
        if (_lodash.default.isNull(tags[tag])) {
          delete newData[_core.DIST_TAGS][tag];
          continue;
        }
        if (_lodash.default.isNil(newData.versions[tags[tag]])) {
          throw _core.errorUtils.getNotFound(_core.API_ERROR.VERSION_NOT_EXIST);
        }
        const version = tags[tag];
        newData = (0, _2.tagVersionNext)(newData, version, tag);
      }
      return newData;
    });
  }
  getUpLinkForDistFile(pkgName, distFile) {
    let uplink = null;
    for (const uplinkId in this.uplinks) {
      // refer to https://github.com/verdaccio/verdaccio/issues/1642
      if ((0, _config.hasProxyTo)(pkgName, uplinkId, this.config.packages)) {
        uplink = this.uplinks[uplinkId];
      }
    }
    if (uplink == null) {
      debug('upstream not found creating one for %o', pkgName);
      uplink = new _proxy.ProxyStorage({
        url: distFile.url,
        cache: true
      }, this.config, this.logger);
    }
    return uplink;
  }
  async updateManifest(manifest, options) {
    debug('update manifest %o for user %o', manifest._id, options.requestOptions.username);
    if ((0, _2.isDeprecatedManifest)(manifest)) {
      // if the manifest is deprecated, we need to update the package.json
      await this.deprecate(manifest, {
        ...options
      });
    } else if ((0, _starUtils.isPublishablePackage)(manifest) === false && _core.validatioUtils.isObject(manifest.users)) {
      // if user request to apply a star to the manifest
      await this.star(manifest, {
        ...options
      });
      return _core.API_MESSAGE.PKG_CHANGED;
    } else if ((0, _starUtils.isPublishablePackage)(manifest) === false && Array.isArray(manifest.maintainers)) {
      // if user request to change owners of package
      await this.changeOwners(manifest, {
        ...options
      });
      return _core.API_MESSAGE.PKG_CHANGED;
    } else if (_core.validatioUtils.validatePublishSingleVersion(manifest)) {
      // if continue, the version to be published does not exist
      // we create a new package
      const [mergedManifest, version, message] = await this.publishANewVersion(manifest, {
        ...options
      });
      // send notification of publication (notification step, non transactional)
      try {
        const {
          name
        } = mergedManifest;
        await this.notify(mergedManifest, `${name}@${version}`);
        this.logger.info({
          name,
          version
        }, 'notify for @{name}@@{version} has been sent');
      } catch (error) {
        this.logger.error({
          error: error.message
        }, 'notify batch service has failed: @{error}');
      }
      return message;
    } else {
      debug('invalid body format');
      this.logger.warn({
        packageName: options.name
      }, `wrong package format on publish a package @{packageName}`);
      throw _core.errorUtils.getBadRequest(_core.API_ERROR.UNSUPORTED_REGISTRY_CALL);
    }
  }
  async deprecate(manifest, options) {
    const {
      name
    } = manifest;
    debug('deprecating %s', name);
    return this.changePackage(name, manifest, options.revision);
  }
  async star(manifest, options) {
    const {
      users
    } = manifest;
    const {
      requestOptions,
      name
    } = options;
    debug('star %s', name);
    const {
      username
    } = requestOptions;
    if (!username) {
      throw _core.errorUtils.getBadRequest('update users only allowed to logged users');
    }
    const localPackage = await this.getPackageManifest({
      name,
      requestOptions,
      uplinksLook: false
    });
    // backward compatible in case users are not in the storage.
    const localStarUsers = localPackage.users || {};
    // if trying to add a star
    const userIsAddingStar = Object.keys(users).includes(username);
    if (!(0, _starUtils.isExecutingStarCommand)(localPackage?.users, username, userIsAddingStar)) {
      return _core.API_MESSAGE.PKG_CHANGED;
    }
    const newUsers = userIsAddingStar ? {
      ...localStarUsers,
      [username]: true
    } : _lodash.default.reduce(localStarUsers, (users, value, key) => {
      if (key !== username) {
        users[key] = value;
      }
      return users;
    }, {});
    await this.changePackage(name, {
      ...localPackage,
      users: newUsers
    }, options.revision);
    return _core.API_MESSAGE.PKG_CHANGED;
  }
  async changeOwners(manifest, options) {
    const {
      maintainers
    } = manifest;
    const {
      requestOptions,
      name
    } = options;
    debug('change owners of %o', name);
    const {
      username
    } = requestOptions;
    if (!username) {
      throw _core.errorUtils.getBadRequest('update owners only allowed for logged in users');
    }
    if (!maintainers || maintainers.length === 0) {
      throw _core.errorUtils.getBadRequest('maintainers field is required and must not be empty');
    }
    const localPackage = await this.getPackageManifest({
      name,
      requestOptions,
      uplinksLook: false
    });
    await this.changePackage(name, {
      ...localPackage,
      maintainers: maintainers
    }, options.revision);
    return _core.API_MESSAGE.PKG_CHANGED;
  }

  /**
   * Get local package, on fails return null.
   * Errors are considered package not found.
   * @param name
   * @returns
   */
  async getPackagelocalByNameNext(name) {
    try {
      return await this.getPackageLocalMetadata(name);
    } catch (err) {
      debug('local package %s not found', name);
      return null;
    }
  }

  /**
   * Convert tarball as string into a Buffer and validate the length.
   * @param data the tarball data as string
   * @returns
   */
  getBufferManifest(data) {
    const buffer = Buffer.from(data, 'base64');
    if (buffer.length === 0) {
      throw _core.errorUtils.getBadData('refusing to accept zero-length file');
    }
    return buffer;
  }

  /**
   * Verify if the package exists in the local storage
   * (the package refers to the package.json), directory would return false.
   * @param pkgName package name
   * @returns boolean
   */
  async hasPackage(pkgName) {
    const storage = this.getPrivatePackageStorage(pkgName);
    if (typeof storage === 'undefined') {
      throw _core.errorUtils.getNotFound();
    }
    const hasPackage = await storage.hasPackage();
    debug('has package %o for %o', pkgName, hasPackage);
    return hasPackage;
  }

  /**
   * Create a new package.
   * This situation happens only of the package does not exist on the cache.
   *
   * @param body package metadata
   * @param options
   * @returns
   */
  async publishANewVersion(body, options) {
    const {
      name
    } = options;
    const username = options.requestOptions.username;
    debug('publishing a new package for %o as %o', name, username);
    let successResponseMessage;
    const manifest = {
      ..._core.validatioUtils.normalizeMetadata(body, name)
    };
    const {
      _attachments,
      versions
    } = manifest;

    // validation step, if _attachments is not an object throw error
    // _attachments is need it for holding the tarball buffer in the local storage as file
    // versions is need it for holding the version in the local storage as file
    // _attachments and validation are required otherwise cannot continue.
    if ((0, _lodash.isEmpty)(_attachments)) {
      throw _core.errorUtils.getBadRequest(_core.API_ERROR.UNSUPORTED_REGISTRY_CALL);
    }

    // get the unique version available
    const [versionToPublish] = Object.keys(versions);

    // at this point document is either created or existed before
    const [firstAttachmentKey] = Object.keys(_attachments);
    const buffer = this.getBufferManifest(body._attachments[firstAttachmentKey].data);
    try {
      // we check if package exist already locally
      const localManifest = await this.getPackagelocalByNameNext(name);
      // if continue, the version to be published does not exist
      if (localManifest?.versions[versionToPublish] != null) {
        debug('%s version %s already exists (locally)', name, versionToPublish);
        throw _core.errorUtils.getConflict();
      }
      const uplinksLook = this.config?.publish?.allow_offline === false;
      // if execution get here, package does not exist locally, we search upstream
      const remoteManifest = await this.checkPackageRemote(name, uplinksLook);
      if (remoteManifest?.versions[versionToPublish] != null) {
        debug('%s version %s already exists (upstream)', name, versionToPublish);
        throw _core.errorUtils.getConflict();
      }
      const hasPackageInStorage = await this.hasPackage(name);
      if (!hasPackageInStorage) {
        await this.createNewLocalCachePackage(name, username);
        successResponseMessage = _core.API_MESSAGE.PKG_CREATED;
      } else {
        await this.checkAllowedToChangePackage(localManifest, username);
        successResponseMessage = _core.API_MESSAGE.PKG_CHANGED;
      }
    } catch (err) {
      debug('error on change or update a package with %o', err.message);
      this.logger.error({
        err: err.message
      }, 'error on publish new version: @{err}');
      throw err;
    }

    // 1. after tarball has been successfully uploaded, we update the version
    try {
      const tarballStats = await this.getTarballStats(versions[versionToPublish], buffer);
      // Older package managers like npm6 do not send readme content as part of version but include it on root level
      if (_lodash.default.isEmpty(versions[versionToPublish].readme)) {
        versions[versionToPublish].readme = _lodash.default.isNil(manifest.readme) === false ? String(manifest.readme) : '';
      }
      // addVersion will move the readme from the the published version to the root level
      await this.addVersion(name, versionToPublish, versions[versionToPublish], null, tarballStats);
    } catch (err) {
      this.logger.error({
        err: err.message
      }, 'updated version has failed: @{err}');
      debug('error on create a version for %o with error %o', name, err.message);
      throw err;
    }

    // 2. update and merge tags
    let mergedManifest;
    try {
      // note: I could merge this with addVersion()
      // 1. add version
      // 2. merge versions
      // 3. upload tarball
      // 4. update once to the storage (easy peasy)
      mergedManifest = await this.mergeTagsNext(name, manifest[_core.DIST_TAGS]);
    } catch (err) {
      this.logger.error({
        err: err.message
      }, 'merge version has failed: @{err}');
      debug('error on create a version for %o with error %o', name, err.message);
      // TODO: undo if this fails
      // 1. remove updated version
      throw err;
    }

    // 3. upload the tarball to the storage
    try {
      const readable = _stream.Readable.from(buffer);
      await this.uploadTarball(name, (0, _path.basename)(firstAttachmentKey), readable, {
        signal: options.signal
      });
    } catch (err) {
      this.logger.error({
        err: err.message
      }, 'upload tarball has failed: @{err}');
      // TODO: undo if this fails
      // 1. remove updated version
      // 2. remove new dist tags
      throw err;
    }
    this.logger.info({
      name,
      version: versionToPublish
    }, 'package @{name}@@{version} has been published');
    return [mergedManifest, versionToPublish, successResponseMessage];
  }

  // TODO: pending implementation
  // eslint-disable-next-line @typescript-eslint/no-unused-vars
  async notify(_manifest, _message) {
    return;
  }
  getProxyList() {
    const uplinksList = Object.keys(this.uplinks);
    return uplinksList;
  }

  /**
   * Wrap uploadTarballAsStream into a promise.
   * @param name package name
   * @param fileName tarball name
   * @param contentReadable content as readable stream
   * @param options
   * @returns
   */
  uploadTarball(name, fileName, contentReadable, {
    signal
  }) {
    return new Promise((resolve, reject) => {
      this.uploadTarballAsStream(name, fileName, {
        signal
      }).then(stream => {
        stream.on('error', err => {
          debug('error on stream a tarball %o for %o with error %o', 'foo.tar.gz', name, err.message);
          reject(err);
        });
        stream.on('success', () => {
          this.logger.debug({
            fileName,
            name
          }, 'file @{fileName} for package @{name} has been successfully uploaded');
          resolve();
        });
        return stream;
      }).then(stream => {
        (0, _promises.pipeline)(contentReadable, stream, {
          signal
        }).then(() => {
          debug('success pipe upload tarball');
        }).catch(reject);
      });
    });
  }
  async uploadTarballAsStream(pkgName, filename, {
    signal
  }) {
    debug(`add a tarball for %o`, pkgName);
    (0, _assert.default)(_core.validatioUtils.validateName(filename));
    const shaOneHash = (0, _utils.createTarballHash)();
    const transformHash = new _stream.Transform({
      transform(chunk, _encoding, callback) {
        // measure the length for validation reasons
        shaOneHash.update(chunk);
        callback(null, chunk);
      }
    });
    const uploadStream = new _stream.PassThrough();
    const storage = this.getPrivatePackageStorage(pkgName);

    // FUTURE: this validation could happen even before
    if (pkgName === PROTO_NAME) {
      process.nextTick(() => {
        uploadStream.emit('error', _core.errorUtils.getForbidden());
      });
      return uploadStream;
    }

    // FIXME: this condition will never met, storage is always defined
    if (!storage) {
      process.nextTick(() => {
        uploadStream.emit('error', "can't upload this package storage is missing");
      });
      return uploadStream;
    }
    const fileDoesExist = await storage.hasTarball(filename);
    if (fileDoesExist) {
      process.nextTick(() => {
        uploadStream.emit('error', _core.errorUtils.getConflict());
      });
    } else {
      const localStorageWriteStream = await storage.writeTarball(filename, {
        signal
      });
      localStorageWriteStream.on('open', async () => {
        await (0, _promises.pipeline)(uploadStream, transformHash, localStorageWriteStream, {
          signal
        });
      });

      // once the file descriptor has been closed
      localStorageWriteStream.on('close', async () => {
        try {
          debug('uploaded tarball %o for %o', filename, pkgName);
          // update the package metadata
          await this.updatePackage(pkgName, async data => {
            const newData = {
              ...data
            };
            debug('added _attachment for %o', pkgName);
            newData._attachments[filename] = {
              // TODO:  add integrity hash here
              shasum: shaOneHash.digest('hex')
            };
            return newData;
          });
          debug('emit success for %o', pkgName);
          uploadStream.emit('success');
        } catch (err) {
          // FUTURE: if the update package fails, remove tarball to avoid left
          // orphan tarballs
          debug('something has failed on upload tarball %o for %o : %s', filename, pkgName, err.message);
          uploadStream.emit('error', err);
        }
      });

      // something went wrong writing into the local storage
      localStorageWriteStream.on('error', async err => {
        uploadStream.emit('error', err);
      });
    }
    return uploadStream;
  }

  /**
   * Add a new version to a package.
   * @param name package name
   * @param version version
   * @param metadata version metadata
   * @param tag tag of the version
   */
  async addVersion(name, version, metadata, tag, tarballStats) {
    debug(`add version %s package for %s`, version, name);
    await this.updatePackage(name, async data => {
      // keep only one readme per package
      data.readme = metadata.readme;
      debug('%s readme mutated', name);
      // TODO: lodash remove
      metadata = (0, _2.cleanUpReadme)(metadata);
      metadata.contributors = (0, _utils.normalizeContributors)(metadata.contributors);
      debug('%s contributors normalized', name);

      // Copy current owners to version
      metadata.maintainers = data.maintainers;

      // Update tarball stats
      if (metadata.dist) {
        metadata.dist.fileCount = tarballStats.fileCount;
        metadata.dist.unpackedSize = tarballStats.unpackedSize;
      }

      // if uploaded tarball has a different shasum, it's very likely that we
      // have some kind of error
      if (_core.validatioUtils.isObject(metadata.dist) && _lodash.default.isString(metadata.dist.tarball)) {
        const tarball = (0, _tarball.extractTarballFromUrl)(metadata.dist.tarball);
        if (_core.validatioUtils.isObject(data._attachments[tarball])) {
          if (_lodash.default.isNil(data._attachments[tarball].shasum) === false && _lodash.default.isNil(metadata.dist.shasum) === false) {
            if (data._attachments[tarball].shasum != metadata.dist.shasum) {
              const errorMessage = `shasum error, ` + `${data._attachments[tarball].shasum} != ${metadata.dist.shasum}`;
              throw _core.errorUtils.getBadRequest(errorMessage);
            }
          }
          if (tarball === '__proto__' || tarball === 'constructor' || tarball === 'prototype') {
            throw _core.errorUtils.getBadRequest('tarball name is not allowed');
          }
          data._attachments[tarball].version = version;
        }

        // if the time field doesn't exist, we create it, some old storage
        // might not have it
        // https://github.com/verdaccio/verdaccio/issues/740
        if (_lodash.default.isNil(data.time)) {
          this.logger.warn({
            name
          }, 'time field could not be found, it has been recreated for @{name}');
          data.time = {};
        }

        // populate the time field  with the date of the version
        const currentDate = new Date().toISOString();
        data.time.modified = currentDate;
        if (!data.time?.created) {
          data.time.created = currentDate;
        }
        if (typeof data.time[version] === 'string') {
          // this should not h appen, but we keep this check to avoid or easy bug report
          this.logger.warn({
            name,
            version
          }, 'the time for the version @{version} already exists, it has been overwritten for package @{name}');
        }
        data.time[version] = currentDate;
        debug('time added for %s version %s', name, version);
      }
      data.versions[version] = metadata;
      // TODO: review this method, it's a bit ugly
      (0, _2.tagVersion)(data, version, tag);
      try {
        debug('%s add on database', name);
        await this.localStorage.getStoragePlugin().add(name);
        this.logger.debug({
          name,
          version
        }, 'version @{version} added to database for @{name}');
      } catch (err) {
        throw _core.errorUtils.getBadData(err.message);
      }
      return data;
    });
  }

  /**
   * Create an empty new local cache package without versions.
   * @param name name of the package
   * @returns
   */
  async createNewLocalCachePackage(name, username) {
    const storage = this.getPrivatePackageStorage(name);
    if (!storage) {
      debug(`storage is missing for %o package cannot be added`, name);
      throw _core.errorUtils.getNotFound('this package cannot be added');
    }
    const currentTime = new Date().toISOString();
    const packageData = {
      ...(0, _storageUtils.generatePackageTemplate)(name),
      time: {
        created: currentTime,
        modified: currentTime
      }
    };

    // Set initial package owner
    // TODO: Add email of user
    packageData.maintainers = username && username.length > 0 ? [{
      name: username,
      email: ''
    }] : [{
      name: _core.DEFAULT_USER,
      email: ''
    }];
    try {
      await storage.createPackage(name, packageData);
      this.logger.info({
        name
      }, 'created new package @{name}');
      return;
    } catch (err) {
      if (_lodash.default.isNull(err) === false && (err.code === _storageUtils.STORAGE.FILE_EXIST_ERROR || err.code === _core.HTTP_STATUS.CONFLICT)) {
        debug(`error on creating a package for %o with error %o`, name, err.message);
        throw _core.errorUtils.getConflict();
      }
      return;
    }
  }

  /**
   *
   * @param name package name
   * @param uplinksLook
   * @returns
   */
  async checkPackageRemote(name, uplinksLook) {
    try {
      // we provide a null manifest, thus the manifest returned will be the remote one
      const [remoteManifest, upLinksErrors] = await this.syncUplinksMetadata(name, null, {
        uplinksLook
      });

      // checking package exist already
      if ((0, _lodash.isNil)(remoteManifest) === false) {
        throw _core.errorUtils.getConflict(_core.API_ERROR.PACKAGE_EXIST);
      }
      for (let errorItem = 0; errorItem < upLinksErrors.length; errorItem++) {
        // checking error
        // if uplink fails with a status other than 404, we report failure
        if ((0, _lodash.isNil)(upLinksErrors[errorItem][0]) === false) {
          if (upLinksErrors[errorItem][0].status !== _core.HTTP_STATUS.NOT_FOUND) {
            if (upLinksErrors) {
              return null;
            }
            throw _core.errorUtils.getServiceUnavailable(_core.API_ERROR.UPLINK_OFFLINE_PUBLISH);
          }
        }
      }
      return remoteManifest;
    } catch (err) {
      if (err && err.status !== _core.HTTP_STATUS.NOT_FOUND) {
        throw err;
      }
      return null;
    }
  }
  setDefaultRevision(json) {
    // calculate revision from couch db
    if (_lodash.default.isString(json._rev) === false) {
      json._rev = _storageUtils.STORAGE.DEFAULT_REVISION;
    }

    // this is intended in debug mode we do not want modify the store revision
    if (_lodash.default.isNil(this.config._debug)) {
      const prev = json._rev;
      json._rev = (0, _storageUtils.generateRevision)(json._rev);
      debug('revision metadata for %s updated from %s to %s', json.name, prev, json._rev);
    }
    return json;
  }
  async writePackage(name, json) {
    const storage = this.getPrivatePackageStorage(name);
    if (_lodash.default.isNil(storage)) {
      // TODO: replace here 500 error
      throw _core.errorUtils.getBadData();
    }
    await storage.savePackage(name, this.setDefaultRevision(json));
  }

  /**
   * @param {*} name package name
   * @param {*} updateHandler function(package, cb) - update function
   * @param {*} callback callback that gets invoked after it's all updated
   * @return {Function}
   */
  async updatePackage(name, updateHandler) {
    const storage = this.getPrivatePackageStorage(name);
    if (!storage) {
      throw _core.errorUtils.getNotFound();
    }

    // we update the package on the local storage
    const updatedManifest = await storage.updatePackage(name, updateHandler);
    // after correctly updated write to the storage
    try {
      await this.writePackage(name, (0, _storageUtils.normalizePackage)(updatedManifest));
      return updatedManifest;
    } catch (err) {
      if (err.code === resourceNotAvailable) {
        throw _core.errorUtils.getInternalError('resource temporarily unavailable');
      } else if (err.code === noSuchFile) {
        throw _core.errorUtils.getNotFound();
      } else {
        throw err;
      }
    }
  }

  /**
   *
   * @protected
   * @param {IGetPackageOptionsNext} options
   * @return {*}  {Promise<[Manifest, any[]]>}
   * @memberof AbstractStorage
   */
  async getPackageNext(options) {
    const {
      name
    } = options;
    debug('get package for %o', name);
    let data = null;
    try {
      data = await this.getPackageLocalMetadata(name);
    } catch (err) {
      // if error code is higher than 500 stop here
      if (err && (!err.status || err.status >= _core.HTTP_STATUS.INTERNAL_ERROR)) {
        throw err;
      }
    }

    // if we can't get the local metadata, we try to get the remote metadata
    // if we do to have local metadata, we try to update it with the upstream registry
    debug('sync uplinks for %o', name);
    const [remoteManifest, upLinksErrors] = await this.syncUplinksMetadata(name, data, {
      uplinksLook: options.uplinksLook,
      retry: options.retry,
      remoteAddress: options.requestOptions.remoteAddress
      // etag??
    });

    // if either local data and upstream data are empty, we throw an error
    if (!remoteManifest && _lodash.default.isNull(data)) {
      throw _core.errorUtils.getNotFound(`${_core.API_ERROR.NOT_PACKAGE_UPLINK}: ${name}`);
      // if the remote manifest is empty, we return local data
    } else if (!remoteManifest && !_lodash.default.isNull(data)) {
      // no data on uplinks
      return [data, upLinksErrors];
    }

    // if we have local data, we try to update it with the upstream registry
    const normalizedPkg = Object.assign({}, remoteManifest, {
      // FIXME: clean up  mutation within cleanUpLinksRef method
      ...(0, _storageUtils.normalizeDistTags)((0, _storageUtils.cleanUpLinksRef)(remoteManifest, options.keepUpLinkData)),
      _attachments: {}
    });
    debug('no sync uplinks errors %o for %s', upLinksErrors?.length, name);
    return [normalizedPkg, upLinksErrors];
  }

  /**
   * Function fetches package metadata from uplinks and synchronizes it with local data
     if package is available locally, it MUST be provided in pkginfo.
     Using this example:
     "jquery":
      access: $all
      publish: $authenticated
      unpublish: $authenticated
      # two uplinks setup
      proxy: ver npmjs
      # one uplink setup
      proxy: npmjs
     A package requires uplinks syncronization if enables the proxy section, uplinks
    can be more than one, the more are the most slow request will take, the request
    are made in serial and if 1st call fails, the second will be triggered, otherwise
    the 1st will reply and others will be discarded. The order is important.
     Errors on upkinks are considered are, time outs, connection fails and http status 304,
    in that case the request returns empty body and we want ask next on the list if has fresh
    updates.
   */
  async syncUplinksMetadata(name, localManifest, options = {}) {
    let found = localManifest !== null;
    let syncManifest = null;
    const upLinks = [];
    const hasToLookIntoUplinks = _lodash.default.isNil(options.uplinksLook) || options.uplinksLook;
    debug('is sync uplink enabled %o', hasToLookIntoUplinks);
    for (const uplink in this.uplinks) {
      if ((0, _config.hasProxyTo)(name, uplink, this.config.packages) && hasToLookIntoUplinks) {
        debug('sync uplink %o', uplink);
        upLinks.push(uplink);
      }
    }

    //  if none uplink match we return the local manifest
    if (upLinks.length === 0) {
      debug('no uplinks found for %o upstream update aborted', name);
      return [localManifest, []];
    }
    const uplinksErrors = [];
    // we resolve uplinks async in series, first come first serve
    for (const uplink of upLinks) {
      try {
        const tempManifest = _lodash.default.isNil(localManifest) ? (0, _storageUtils.generatePackageTemplate)(name) : {
          ...localManifest
        };
        syncManifest = await this.mergeCacheRemoteMetadata(this.uplinks[uplink], tempManifest, options);
        debug('syncing on uplink %o', syncManifest.name);
        if (_lodash.default.isNil(syncManifest) === false) {
          found = true;
          break;
        }
      } catch (err) {
        debug('error captured on uplink %o', err.message);
        uplinksErrors.push(err);
        // enforce use next uplink on the list
        continue;
      }
    }
    if (found && syncManifest !== null) {
      // updates the local cache manifest with fresh data
      let updatedCacheManifest = await this.updateVersionsNext(name, syncManifest);
      // plugin filter applied to the manifest
      const [filteredManifest, filtersErrors] = await this.applyFilters(updatedCacheManifest);
      return [{
        ...updatedCacheManifest,
        ...filteredManifest
      }, [...uplinksErrors, ...filtersErrors]];
    } else if (found && _lodash.default.isNil(localManifest) === false) {
      return [localManifest, uplinksErrors];
    } else {
      // if is not found, calculate the right error to return
      debug('uplinks sync failed with %o errors', uplinksErrors.length);
      for (const err of uplinksErrors) {
        const {
          code
        } = err;
        if (code === 'ETIMEDOUT' || code === 'ESOCKETTIMEDOUT' || code === 'ECONNRESET') {
          debug('uplinks sync failed with timeout error');
          throw _core.errorUtils.getServiceUnavailable(err.code);
        }
        // we bubble up the 304 special error case
        if (code === _core.HTTP_STATUS.NOT_MODIFIED) {
          debug('uplinks sync failed with 304 error');
          throw err;
        }
      }
      debug('uplinks sync failed with no package found');
      throw _core.errorUtils.getNotFound(_core.API_ERROR.NO_PACKAGE);
    }
  }

  /**
   * Merge a manifest with a remote manifest.
   *
   * If the uplinks are not available, the local manifest is returned.
   * If the uplinks are available, the local manifest is merged with the remote one.
   *
   *
   * @param uplink uplink instance
   * @param cachedManifest the local cached manifest
   * @param options options
   * @returns Returns a promise that resolves with the merged manifest.
   */
  async mergeCacheRemoteMetadata(uplink, cachedManifest, options) {
    // we store which uplink is updating the manifest
    const upLinkMeta = cachedManifest._uplinks[uplink.upname];
    let _cacheManifest = {
      ...cachedManifest
    };
    if (_core.validatioUtils.isObject(upLinkMeta)) {
      const fetched = upLinkMeta.fetched;

      // we check the uplink cache is fresh
      if (fetched && Date.now() - fetched < uplink.maxage) {
        return cachedManifest;
      }
    }
    const remoteOptions = Object.assign({}, options, {
      etag: upLinkMeta?.etag
    });

    // get the latest metadata from the uplink
    const [remoteManifest, etag] = await uplink.getRemoteMetadata(_cacheManifest.name, remoteOptions);
    try {
      _cacheManifest = _core.validatioUtils.normalizeMetadata(_cacheManifest, _cacheManifest.name);
    } catch (err) {
      this.logger.error({
        err: err
      }, 'package.json validating error @{!err?.message}\n@{err.stack}');
      throw err;
    }
    // updates the _uplink metadata fields, cache, etc
    _cacheManifest = (0, _storageUtils.updateUpLinkMetadata)(uplink.upname, _cacheManifest, etag);
    // merge time field cache and remote
    _cacheManifest = (0, _storageUtils.mergeUplinkTimeIntoLocalNext)(_cacheManifest, remoteManifest);
    // update the _uplinks field in the cache
    _cacheManifest = (0, _proxy.updateVersionsHiddenUpLinkNext)(_cacheManifest, uplink);
    try {
      // merge versions from remote into the cache
      _cacheManifest = (0, _storageUtils.mergeVersions)(_cacheManifest, remoteManifest);
      return _cacheManifest;
    } catch (err) {
      this.logger.error({
        err: err
      }, 'package.json merge has failed @{!err?.message}\n@{err.stack}');
      throw err;
    }
  }

  /**
   * Apply filters to manifest.
   * @param manifest
   * @returns
   */
  async applyFilters(manifest) {
    if (this.filters === null || this.filters.length === 0) {
      return [manifest, []];
    }
    let filterPluginErrors = [];
    let filteredManifest = {
      ...manifest
    };
    for (const filter of this.filters) {
      // These filters can assume it's save to modify packageJsonLocal
      // and return it directly for
      // performance (i.e. need not be pure)
      try {
        filteredManifest = await filter.filter_metadata(manifest);
      } catch (err) {
        this.logger.error({
          err: err.message
        }, 'filter has failed @{err}');
        filterPluginErrors.push(err);
      }
    }
    return [filteredManifest, filterPluginErrors];
  }
  _createNewPackageNext(name) {
    return (0, _storageUtils.normalizePackage)((0, _storageUtils.generatePackageTemplate)(name));
  }

  /**
   * Ensure the dist file remains as the same protocol
   * @param {Object} hash metadata
   * @param {String} upLinkKey registry key
   * @private
   * @deprecated use _updateUplinkToRemoteProtocolNext (???)
   */
  updateUplinkToRemoteProtocol(hash, upLinkKey) {
    // if we got this information from a known registry,
    // use the same protocol for the tarball
    const tarballUrl = _url.default.parse(hash.url);
    const uplinkUrl = _url.default.parse(this.config.uplinks[upLinkKey].url);
    if (uplinkUrl.host === tarballUrl.host) {
      tarballUrl.protocol = uplinkUrl.protocol;
      hash.registry = upLinkKey;
      hash.url = _url.default.format(tarballUrl);
    }
  }

  /**
   * Create or read a package.
   *
   * If the package already exists, it will be read.
   * If the package is not found, it will be created.
   * If the error is anything else will throw an error
   *
   * @param {*} pkgName
   * @param {*} callback
   * @return {Function}
   */
  async readCreatePackage(pkgName) {
    const storage = this.getPrivatePackageStorage(pkgName);
    if (_lodash.default.isNil(storage)) {
      throw _core.errorUtils.getInternalError('storage could not be found');
    }
    try {
      const result = await storage.readPackage(pkgName);
      return (0, _storageUtils.normalizePackage)(result);
    } catch (err) {
      if (err.code === _storageUtils.STORAGE.NO_SUCH_FILE_ERROR || err.code === _core.HTTP_STATUS.NOT_FOUND) {
        return this._createNewPackageNext(pkgName);
      } else {
        this.logger.error({
          err: err,
          file: _storageUtils.STORAGE.PACKAGE_FILE_NAME
        }, `'error reading'  @{file}: @{!err.message}`);
        throw _core.errorUtils.getInternalError();
      }
    }
  }

  /**
    Updates the local cache with the merge from the remote/client manifest.
     The steps are the following.
    1. Get the latest version of the package from the cache.
    2. If does not exist will return a 
     @param name
    @param remoteManifest
    @returns return a merged manifest.
  */
  async updateVersionsNext(name, remoteManifest) {
    debug(`updating versions for package %o`, name);
    let cacheManifest = await this.readCreatePackage(name);
    let change = false;
    // updating readme
    cacheManifest.readme = (0, _storageUtils.getLatestReadme)(remoteManifest);
    if (remoteManifest.readme !== cacheManifest.readme) {
      debug('manifest readme updated for %o', name);
      change = true;
    }
    debug('updating new remote versions');
    for (const versionId in remoteManifest.versions) {
      // if detect a new remote version does not exist cache
      if (_lodash.default.isNil(cacheManifest.versions[versionId])) {
        debug('new version from upstream %o', versionId);
        let version = remoteManifest.versions[versionId];

        // we don't keep readme for package versions,
        // only one readme per package
        // TODO: readme clean up could be  saved in configured eventually
        version = (0, _2.cleanUpReadme)(version);
        debug('clean up readme for %o', versionId);
        version.contributors = (0, _utils.normalizeContributors)(version.contributors);
        change = true;
        cacheManifest.versions[versionId] = version;
        if (version?.dist?.tarball) {
          const filename = _core.pkgUtils.extractTarballName(version.dist.tarball);
          // store a fast access to the dist file by tarball name
          // it does NOT overwrite any existing records
          if (_lodash.default.isNil(cacheManifest?._distfiles[filename])) {
            const hash = cacheManifest._distfiles[filename] = {
              url: version.dist.tarball,
              sha: version.dist.shasum
            };
            // store cache metadata this the manifest
            const upLink = version[Symbol.for('__verdaccio_uplink')];
            if (_lodash.default.isNil(upLink) === false) {
              this.updateUplinkToRemoteProtocol(hash, upLink);
            }
          }
        }
      } else {
        debug('no new versions from upstream %s', name);
      }
    }
    debug('update dist-tags');
    for (const tag in remoteManifest[_core.DIST_TAGS]) {
      if (!cacheManifest[_core.DIST_TAGS][tag] || cacheManifest[_core.DIST_TAGS][tag] !== remoteManifest[_core.DIST_TAGS][tag]) {
        change = true;
        cacheManifest[_core.DIST_TAGS][tag] = remoteManifest[_core.DIST_TAGS][tag];
      }
    }
    for (const up in remoteManifest._uplinks) {
      if (Object.prototype.hasOwnProperty.call(remoteManifest._uplinks, up)) {
        const need_change = !(0, _utils.isObject)(cacheManifest._uplinks[up]) || remoteManifest._uplinks[up].etag !== cacheManifest._uplinks[up].etag || remoteManifest._uplinks[up].fetched !== cacheManifest._uplinks[up].fetched;
        if (need_change) {
          change = true;
          cacheManifest._uplinks[up] = remoteManifest._uplinks[up];
        }
      }
    }
    debug('update time');
    if ('time' in remoteManifest && !_lodash.default.isEqual(cacheManifest.time, remoteManifest.time)) {
      cacheManifest.time = remoteManifest.time;
      change = true;
    }
    if (change) {
      debug('updating package info %o', name);
      await this.writePackage(name, cacheManifest);
      return cacheManifest;
    } else {
      return cacheManifest;
    }
  }
  async getTarballStats(version, buffer) {
    if (version.dist == undefined || version.dist?.fileCount == undefined || version.dist?.unpackedSize == undefined) {
      debug('tarball stats not found, calculating');
      return await (0, _tarball.getTarballDetails)(buffer);
    } else {
      debug('tarball stats found');
      return {
        fileCount: version.dist.fileCount,
        unpackedSize: version.dist.unpackedSize
      };
    }
  }
  async checkAllowedToChangePackage(manifest, username) {
    // Checks to perform if config "publish:check_owners" is true
    debug('check if user %o is an owner and allowed to change package', username);
    // if name of owner is not included in list of maintainers, then throw an error
    if (this.config?.publish?.check_owners === true && manifest.maintainers && manifest.maintainers.length > 0 && !manifest.maintainers.some(maintainer => maintainer.name === username)) {
      this.logger.error({
        username
      }, '@{username} is not a maintainer (package owner)');
      throw Error('only owners are allowed to change package');
    }
  }
}
exports.Storage = Storage;
//# sourceMappingURL=storage.js.map